/* -*- c-file-style: "linux" -*-
 * ssl.c: operations for negotiating SSL rsync connections.
 *
 * Copyright (C) 2003  Casey Marshall <rsdio@metastatic.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "rsync.h"

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#else
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#endif
#include <string.h>

#define SSL_MAX_RECORD_SIZE (1024*16) /* TLS record size */
#define MAX_BUFFERING_TIME_MS	100

extern int verbose;
extern int am_daemon;
extern int am_server;

extern char *ssl_cert_path;
extern char *ssl_key_path;
extern char *ssl_key_passwd;
extern char *ssl_ca_path;

static SSL_CTX *ssl_ctx;
static SSL *ssl;
static int tls_read[2] = { -1, -1 };
static int tls_write[2] = { -1, -1 };
static int ssl_running;
static int ssl_min_send_size;
static int ssl_pid = -1;

#ifdef HAVE_SIGACTION
static struct sigaction sigact;
#endif

/* copied from progress.c */
static unsigned long msdiff(struct timeval *t1, struct timeval *t2)
{
	return (t2->tv_sec - t1->tv_sec) * 1000L
		+ (t2->tv_usec - t1->tv_usec) / 1000;
}

/**
 * A non-interactive callback to be passed to SSL_CTX_set_default_password_cb,
 * which merely copies the value of ssl_key_passwd into buf. This is
 * used for when the private key password is supplied via an option.
 */
static int default_password_cb(char *buf, int n, UNUSED(int f), UNUSED(void *u))
{
	if (ssl_key_passwd == NULL || n < (int)strlen(ssl_key_passwd))
		return 0;
	strncpy(buf, ssl_key_passwd, n-1);
	return strlen(ssl_key_passwd);
}

/**
 * If verbose, this method traces the status of the SSL handshake.
 */
static void info_callback(const SSL *ssl, int cb, int val)
{
	char buf[128];
	char *cbs;

	switch (cb) {
	case SSL_CB_LOOP:
		cbs = "SSL_CB_LOOP";
		break;
	case SSL_CB_EXIT:
		cbs = "SSL_CB_EXIT";
		break;
	case SSL_CB_READ:
		cbs = "SSL_CB_READ";
		break;
	case SSL_CB_WRITE:
		cbs = "SSL_CB_WRITE";
		break;
	case SSL_CB_ALERT:
		cbs = "SSL_CB_ALERT";
		break;
	case SSL_CB_READ_ALERT:
		cbs = "SSL_CB_READ_ALERT";
		break;
	case SSL_CB_WRITE_ALERT:
		cbs = "SSL_CB_WRITE_ALERT";
		break;
	case SSL_CB_ACCEPT_LOOP:
		cbs = "SSL_CB_ACCEPT_LOOP";
		break;
	case SSL_CB_ACCEPT_EXIT:
		cbs = "SSL_CB_ACCEPT_EXIT";
		break;
	case SSL_CB_CONNECT_LOOP:
		cbs = "SSL_CB_CONNECT_LOOP";
		break;
	case SSL_CB_CONNECT_EXIT:
		cbs = "SSL_CB_CONNECT_EXIT";
		break;
	case SSL_CB_HANDSHAKE_START:
		cbs = "SSL_CB_HANDSHAKE_START";
		break;
	case SSL_CB_HANDSHAKE_DONE:
		cbs = "SSL_CB_HANDSHAKE_DONE";
		break;
	default:
		snprintf(buf, sizeof buf, "??? (%d)", cb);
		cbs = buf;
		break;
	}
	if (verbose > 2) {
		rprintf(FLOG, "SSL: info_callback(%p,%s,%d)\n", ssl, cbs, val);
		if (cb == SSL_CB_HANDSHAKE_DONE) {
			SSL_CIPHER_description(SSL_get_current_cipher((SSL*)ssl),
					       buf, sizeof buf);
			rprintf(FLOG, "SSL: cipher: %s", buf);
		}
	}
}

/**
 * Initializes the SSL context for TLSv1 connections; returns zero on
 * success.
 */
int init_tls(void)
{
	if (ssl_ctx)
		return 0;
	SSL_library_init();
	SSL_load_error_strings();
	ssl_ctx = SSL_CTX_new(TLSv1_method());
	if (!ssl_ctx)
		return 1;
	SSL_CTX_set_info_callback(ssl_ctx, info_callback);

	SSL_CTX_set_options(ssl_ctx, SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS|SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3);

	/* Sets the certificate sent to the other party. */
	if (ssl_cert_path != NULL
	    && SSL_CTX_use_certificate_file(ssl_ctx, ssl_cert_path,
					    SSL_FILETYPE_PEM) != 1)
		return 1;
	/* Set up the simple non-interactive callback if the password
	 * was supplied on the command line. */
	if (ssl_key_passwd != NULL)
		SSL_CTX_set_default_passwd_cb(ssl_ctx, default_password_cb);
	/* Sets the private key that matches the public certificate. */
	if (ssl_key_path != NULL) {
		if (SSL_CTX_use_PrivateKey_file(ssl_ctx, ssl_key_path,
						SSL_FILETYPE_PEM) != 1)
			return 1;
		if (SSL_CTX_check_private_key(ssl_ctx) != 1)
			return 1;
	}
	if (ssl_ca_path != NULL
	    && !SSL_CTX_load_verify_locations(ssl_ctx, ssl_ca_path, NULL))
		return 1;

	return 0;
}

/**
 * Returns the error string for the current SSL error, if any.
 */
char *get_ssl_error(void)
{
	return ERR_error_string(ERR_get_error(), NULL);
}

/**
 * Returns the input file descriptor for the SSL connection.
 */
int get_tls_rfd(void)
{
	return tls_read[0];
}

/**
 * Returns the output file descriptor for the SSL connection.
 */
int get_tls_wfd(void)
{
	return tls_write[1];
}

/**
 * Signal handler that ends the SSL connection.
 */
static RETSIGTYPE tls_sigusr1(int UNUSED(val))
{
	ssl_running = 0;
}

/* Signal handler that starts record-size buffering. */
static RETSIGTYPE tls_sigusr2(int UNUSED(val))
{
	ssl_min_send_size = SSL_MAX_RECORD_SIZE;
}

/**
 * Negotiates the TLS connection, creates a socket pair for communicating
 * with the rsync process, then forks into a new process that will handle
 * the communication.
 *
 * 0 is returned on success.
 * rsync to network algorithm:
 * A buffer ring of two buffers is maintained. Data is read from the local
 * socket into the current buffer. The buffer is eligible for sending if
 * enough data has been buffered, or enough time has passed. Once the buffer
 * becomes eligible for sending it is marked "ready" and buffering continues
 * on the other buffer (assuming it's "unready"). This algorithm tries to
 * maximize SSL record size.
 *
 * network to rsync algorithm:
 * One receive buffer twice the size of the SSL max record size. Reading
 * from the network is attempted only if there's enough space for a whole
 * record (this is done to optimize OpenSSL per-call overhead).
 * As bytes enter the buffer sending to rsync begins, when the buffer becomes
 * full we wait for sending to rsync to finish.
 *
 * A graceful stop is attempted on both directions. So if the pipe from rsync
 * is gracefully shut down, we try to send all data to the network before closing
 * the connection and if the SSL connection is gracefully shut down we try to
 * send everything to rsync before closing the rsync pipe.
 */
int start_tls(int f_in, int f_out)
{
	int n = 0, r;
	unsigned char buf_tonetwork[2][SSL_MAX_RECORD_SIZE], buf_fromnetwork[2*SSL_MAX_RECORD_SIZE];
	int tonet_ready[2] = {0}, tonet_size[2] = {0}, write_tonet = 0, tonet_fill_idx = 0, tonet_send_idx = 0;
	struct timeval buffering_start, now, timeout, *tp;
	int avail_fromnet = 0, write_fromnet = 0;
	int want_read_fromnet = 1, want_write_tonet = 0;
	int want_read_fromrsync = 1, want_write_torsync = 0;
	int write_tonet_blocks = 0, read_fromnet_blocks = 0;
	int write_torsync_blocks = 0, read_fromrsync_blocks = 0;
	int ssl_readfd_blocks = 0, ssl_writefd_blocks = 0;
	int ssl_write_count = 0;
	int want_more_io;
	int net_closed = 0, rsync_closed = 0;
	int loopcount;
	fd_set rd, wd;
	const char *ssl_ciphers;

	if (fd_pair(tls_read))
		return 1;
	if (fd_pair(tls_write))
		return 1;

	set_blocking(tls_read[0]);
	set_nonblocking(tls_read[1]);
	set_nonblocking(tls_write[0]);
	set_blocking(tls_write[1]);
	set_nonblocking(f_in);
	set_nonblocking(f_out);

	ssl_pid = do_fork();
	if (ssl_pid < 0)
		return -1;
	if (ssl_pid != 0) {
		close(tls_write[0]);
		close(tls_read[1]);
		close(f_in);
		close(f_out);
		return 0;
	}

	close(tls_write[1]);
	close(tls_read[0]);

	SIGACTION(SIGUSR1, tls_sigusr1);
	SIGACTION(SIGUSR2, tls_sigusr2);
	ssl = SSL_new(ssl_ctx);
	if (!ssl)
		goto out;
	if (am_daemon || am_server)
		SSL_set_accept_state(ssl);
	else {
		SSL_set_connect_state(ssl);
		ssl_ciphers = getenv("RSYNC_SSL_CIPHERS");
		if (ssl_ciphers)
			SSL_set_cipher_list(ssl, ssl_ciphers);
	}

	SSL_set_rfd(ssl, f_in);
	SSL_set_wfd(ssl, f_out);

	n = tls_write[0];
	n = MAX(tls_read[1], n);
	n = MAX(f_in, n);
	n = MAX(f_out, n) + 1;

	ssl_running = 1;
	loopcount = 0;
	while (ssl_running) {
		++loopcount; /* the loopcount prevents starvation of one transfer direction by the other */
		want_more_io = 0;

		if (want_read_fromrsync && !read_fromrsync_blocks) {
			int cur_buf_size = tonet_size[tonet_fill_idx];
			r = read(tls_write[0], buf_tonetwork[tonet_fill_idx]+cur_buf_size, sizeof(buf_tonetwork[0])-cur_buf_size);
			if (r > 0) {
				want_more_io = 1;
				if (!net_closed) {
					if (cur_buf_size == 0)
						gettimeofday(&buffering_start, NULL);
					cur_buf_size += r;
					tonet_size[tonet_fill_idx] = cur_buf_size;
					if (cur_buf_size >= ssl_min_send_size) {
						want_write_tonet = 1;
						tonet_ready[tonet_fill_idx] = 1;
						tonet_fill_idx = !tonet_fill_idx;
						if (tonet_ready[tonet_fill_idx])
							want_read_fromrsync = 0;
					}
				}
			} else if (r < 0 && errno == EWOULDBLOCK)
				read_fromrsync_blocks = 1;
			else if (r < 0) {
				rprintf(FERROR, "pipe read error: %s\n",
					strerror(errno));
				break;
			} else {
				rsync_closed = 1;
				if (cur_buf_size) {
					want_write_tonet = 1;
					tonet_ready[tonet_fill_idx] = 1; /* close current buffer */
				}
				want_read_fromrsync = 0; /* don't read */
				want_write_torsync = 0; /* don't write */
				read_fromrsync_blocks = 0; /* don't select */
				write_torsync_blocks = 0;
				want_read_fromnet = 1; /* purge incoming data from network */
				avail_fromnet = 0;
				if (net_closed || !want_write_tonet)
					break;
			}
		}

		if (want_read_fromnet && !read_fromnet_blocks) {
			r = SSL_read(ssl, buf_fromnetwork+avail_fromnet, SSL_MAX_RECORD_SIZE);
			if (r > 0) {
				want_more_io = 1;
				if (!rsync_closed) {
					avail_fromnet += r;
					want_write_torsync = 1;
					if (avail_fromnet >= (int)(sizeof(buf_fromnetwork)-SSL_MAX_RECORD_SIZE))
						want_read_fromnet = 0;
				}
			} else {
				switch (SSL_get_error(ssl, r)) {
				case SSL_ERROR_ZERO_RETURN:
					net_closed = 1;
					want_read_fromnet = 0; /* don't read */
					want_write_tonet = 0; /* don't write */
					ssl_readfd_blocks = 0; /* and for heaven's sake - don't select() */
					ssl_writefd_blocks = 0;
					want_read_fromrsync = 1; /* and purge stuff from rsync side */
					tonet_size[tonet_fill_idx] = 0;
					tonet_ready[tonet_fill_idx] = 0;
					if (rsync_closed || !want_write_torsync)
						goto graceful_stop;
					break;
				case SSL_ERROR_WANT_READ:
					ssl_readfd_blocks = 1;
					read_fromnet_blocks = 1;
					break;
				case SSL_ERROR_WANT_WRITE:
					ssl_writefd_blocks = 1;
					read_fromnet_blocks = 1;
					break;
				case SSL_ERROR_SYSCALL:
					if (r == 0)
						rprintf(FERROR, "SSL spurious EOF\n");
					else
						rprintf(FERROR, "SSL I/O error: %s\n",
							strerror(errno));
					goto closed;
				case SSL_ERROR_SSL:
					rprintf(FERROR, "SSL %s\n",
						ERR_error_string(ERR_get_error(), NULL));
					goto closed;
				default:
					rprintf(FERROR, "unexpected ssl error %d\n", r);
					goto closed;
				}
			}
		}

		if (want_write_torsync && !write_torsync_blocks) {
			r = write(tls_read[1], buf_fromnetwork+write_fromnet, avail_fromnet-write_fromnet);
			if (r > 0) {
				want_more_io = 1;
				write_fromnet += r;
				if (write_fromnet >= avail_fromnet) {
					write_fromnet = 0;
					avail_fromnet = 0;
					if (net_closed)
						break;
					want_read_fromnet = 1;
					want_write_torsync = 0;
				}
			}
			else if (errno == EWOULDBLOCK)
				write_torsync_blocks = 1;
			else {
				if (errno != EPIPE)
					rprintf(FERROR, "pipe write error: %s\n", strerror(errno));
				break;
			}
		}

		if (want_write_tonet && !write_tonet_blocks) {
			/* lock the write count, 'cause if SSL_write fails on non-blocking IO, it
			 * must be repeated with same parameters. */
			if (ssl_write_count == 0)
				ssl_write_count = tonet_size[tonet_send_idx] - write_tonet;
			r = SSL_write(ssl, buf_tonetwork[tonet_send_idx]+write_tonet, ssl_write_count);
			if (r > 0) {
				want_more_io = 1;
				write_tonet += r;
				ssl_write_count = 0;
				if (write_tonet >= tonet_size[tonet_send_idx]) {
					write_tonet = 0;
					tonet_size[tonet_send_idx] = 0;
					tonet_ready[tonet_send_idx] = 0;
					if (!rsync_closed)
						want_read_fromrsync = 1;
					tonet_send_idx = !tonet_send_idx;
					if (!tonet_ready[tonet_send_idx]) {
						want_write_tonet = 0;
						if (rsync_closed)
							break;
					}
				}
			} else {
				switch (SSL_get_error(ssl, r)) {
				case SSL_ERROR_ZERO_RETURN:
					goto graceful_stop;
				case SSL_ERROR_WANT_READ:
					ssl_readfd_blocks = 1;
					write_tonet_blocks = 1;
					break;
				case SSL_ERROR_WANT_WRITE:
					ssl_writefd_blocks = 1;
					write_tonet_blocks = 1;
					break;
				case SSL_ERROR_SYSCALL:
					if (r == 0)
						rprintf(FERROR, "SSL: spurious EOF\n");
					else
						rprintf(FERROR, "SSL: I/O error: %s\n", strerror(errno));
					goto closed;
				case SSL_ERROR_SSL:
					rprintf(FERROR, "SSL: %s\n",
						ERR_error_string(ERR_get_error(), NULL));
					goto closed;
				default:
					rprintf(FERROR, "unexpected ssl error %d\n", r);
					goto closed;
				}
			}
		}

		if (!want_more_io || (loopcount&31) == 0) {
			int has_buffered = 0;
			if (!net_closed && !tonet_ready[tonet_fill_idx] && tonet_size[tonet_fill_idx]) {
				has_buffered = 1;
				unsigned long buffering_time;
				gettimeofday(&now,NULL);
				buffering_time = msdiff(&buffering_start,&now);
				if (buffering_time >= MAX_BUFFERING_TIME_MS) {
					want_write_tonet = 1;
					want_more_io = 1;
					tonet_ready[tonet_fill_idx] = 1;
					tonet_fill_idx = !tonet_fill_idx;
					if (tonet_ready[tonet_fill_idx])
						want_read_fromrsync = 0;
				}
			}

			int waiting_on_something = 0;
			FD_ZERO(&rd);
			FD_ZERO(&wd);
			if (read_fromrsync_blocks) {
				waiting_on_something = 1;
				FD_SET(tls_write[0], &rd);
			}
			if (write_torsync_blocks) {
				waiting_on_something = 1;
				FD_SET(tls_read[1], &wd);
			}
			if (ssl_readfd_blocks) {
				waiting_on_something = 1;
				FD_SET(f_in, &rd);
			}
			if (ssl_writefd_blocks) {
				waiting_on_something = 1;
				FD_SET(f_out, &wd);
			}

			if (want_more_io) {
				/* just poll to see if some sockets became non-blocking*/
				timeout.tv_sec = 0;
				timeout.tv_usec = 0;
				tp = &timeout;
			} else if (waiting_on_something && !has_buffered)
				tp = NULL; /*infinite wait until a socket becomes available*/
			else {
				timeout.tv_sec = MAX_BUFFERING_TIME_MS/1000;
				timeout.tv_usec = (MAX_BUFFERING_TIME_MS%1000)*1000;
				tp = &timeout;
			}

			if (waiting_on_something || !want_more_io) {
				r = select(n, &rd, &wd, NULL, tp);

				if (r == -1) {
					if (errno != EINTR) {
						rprintf(FERROR, "select error: %s\n",
							strerror(errno));
						break;
					}
				}

				if (r > 0) {
					if (FD_ISSET(tls_write[0], &rd))
						read_fromrsync_blocks = 0;
					if (FD_ISSET(f_in,&rd)) {
						read_fromnet_blocks = 0;
						write_tonet_blocks = 0;
						ssl_readfd_blocks = 0;
					}
					if (FD_ISSET(tls_read[1],&wd))
						write_torsync_blocks = 0;
					if (FD_ISSET(f_out,&wd)) {
						read_fromnet_blocks = 0;
						write_tonet_blocks = 0;
						ssl_writefd_blocks = 0;
					}
				}
			}
		}
	}

  graceful_stop:
	/* The SSL shutdown API looks pretty broken with respect to corner cases
	 * where the underlying socket is blocking. This is a "best effort". */
	if (SSL_shutdown(ssl) != 1)
		SSL_shutdown(ssl);
  closed:
	SSL_free(ssl);

	/* We're finished. */
  out:
	close(tls_read[1]);
	close(tls_write[0]);
	close(f_in);
	if (f_out != f_in)
		close(f_out);
	_exit(0);
}

void start_tls_buffering(void)
{
	if (ssl_pid > 0)
		kill(ssl_pid, SIGUSR2);
}

/**
 * Ends the TLS connection.
 */
void end_tls(void)
{
	if (ssl_pid > 0) {
		int status;
		/* try a graceful stop - this makes sure all data is sent, and
		 * also causes our side to send a close_notify alert, as required
		 * by TLS specs. */
		close(get_tls_rfd());
		if (get_tls_wfd() != get_tls_rfd())
			close(get_tls_wfd());
		terminate_process(ssl_pid, &status, SIGUSR1, 1);
		ssl_pid = -1;
	}
}
